/* tfpwm.c
	Three Phase Sine Generator	
	Uros Platise (c) 2003, <uros.platise@ijs.si>
*/

#include <tfpwmapi.h>
#include "tfpwm.h"

/* to provide debug sync. output so you may see the center-based pulses on the scope */
//#define DEBUG

#define PSoC_Designer_V4

/* fix the diferences between the PSoC builder version 3 and 4 */
#ifdef PSoC_Designer_V4
  #define PWM8_1_PWIDTH_REG	PWM8_1_COMPARE_REG
  #define PWM8_2_PWIDTH_REG	PWM8_2_COMPARE_REG
  #define PWM8_3_PWIDTH_REG	PWM8_3_COMPARE_REG
#endif

/* Table of Half-Wave Symetric Sinus Function */
const char sin_table[SIN_TABLE_N2] = 
{
	0, 2, 4, 6, 8, 10, 12, 14, 16, 18, 20, 22, 24, 26, 28, 
	29, 31, 33, 34, 36, 37, 39, 40, 41, 42, 43, 44, 45, 46, 
	47, 48, 48, 49, 49, 49, 50, 50, 50, 50, 50, 50, 49, 49, 
	49, 48, 48, 47, 46, 45, 44, 43, 42, 41, 40, 39, 37, 36, 
	34, 33, 31, 29, 28, 26, 24, 22, 20, 18, 16, 14, 12, 10, 
	8, 6, 4, 2
};


/* global variables for synchronous load */
unsigned char pwm1, period1, pwm2, period2, pwm3, period3;

/* pwm width are delayed for one period */
unsigned char bpwm1, bpwm2, bpwm3;

/* semaphore */
unsigned char pwm_load_flag = 0;

/* test sync. output */
unsigned char debug_sync_out = 0;


/*
	Sine Retrieve function
*/

unsigned char sine_calcwidth(unsigned char phase, unsigned char amplitude)
{
	unsigned char tval;

	if (phase >= SIN_TABLE_N) phase -= SIN_TABLE_N;
	
	if (phase < SIN_TABLE_N2)
	{
		tval = sin_table[phase];
		return PWM_HPERIOD + (tval << 1);	// make it even, add a zero on right side
	}
	else
	{
		tval = sin_table[phase - SIN_TABLE_N2];
		return PWM_HPERIOD - (tval << 1);	// make it even, add a zero on right side
	}
}




/* 
	Three Phase PWM Fast Interrupt
	
	Note: C-compiler compiles the following ISR so good, 
	as it were almost written in assembler. Minor optimizations
	may still be applied. Check the output .lst file for details.
	
	Placement of the PWM8 blocks _must_ be such that they are 
	assigned highest priority interrupts - all three!
	
	Approximate cycle analyzation:
	
	IRQ_COUNT	PWM_UPLOAD	CYCLES
	---------------------------------------	
	0			-			27+35+43 = 105
	1			-			27+43    = 70
	2			YES			27+73+43 = 143
	2			NO			27+24+43 = 94
	---------------------------------------
	WORST CASE TOTAL: 318 
	
	Details:
	---------------------------------------
	IRQ Entry: 27
	IRQ_Count_0: 35
	IRQ_Count_1: 0
	IRQ_Count_2: 24 / 73
	IRQ_Exit: 43
*/

#pragma interrupt_handler tfpwm_isr

void tfpwm_isr()
{
	static unsigned char irq_count=0;

	if (irq_count==0)
	{				
		PWM8_1_PWIDTH_REG = bpwm1;
		PWM8_2_PWIDTH_REG = bpwm2;
		PWM8_3_PWIDTH_REG = bpwm3;
	}
	else if (irq_count==2)
	{
		if (pwm_load_flag)
		{
			PWM8_1_PERIOD_REG = period1;
			PWM8_2_PERIOD_REG = period2;
			PWM8_3_PERIOD_REG = period3;
			bpwm1 = pwm1;
			bpwm2 = pwm2;
			bpwm3 = pwm3;
#ifdef DEBUG			
			PRT1DR = debug_sync_out;
#endif
			pwm_load_flag = 0;
		}
		else
		{
			/* if load was unsuccessfull, next period has no change! */
			PWM8_1_PERIOD_REG = PWM_PERIOD;
			PWM8_2_PERIOD_REG = PWM_PERIOD;
			PWM8_3_PERIOD_REG = PWM_PERIOD;
		}
	}	
	
	irq_count++;
	if (irq_count==3) irq_count=0;
#ifdef DEBUG
	PRT1DR = 0;
#endif
}


/*
	Three Phase API
*/


void tfpwm_start(unsigned char sine_phase, unsigned char amplitude)
{
	M8C_DisableGInt;

	/* Set comparator bus to low (default) to perform sync. start */
    CMPPRG_1_Start(CMPPRG_1_LOWPOWER);
    CMPPRG_1_SetRef(CMPPRG_1_REF1_000);

	/* Set PWM start values / period and offset
	   Default counter value is updated while counter is stopped.
	*/

	pwm1 = bpwm1 = sine_calcwidth(sine_phase, amplitude);
	PWM8_1_WritePeriod((bpwm1>>1)-1);
    PWM8_1_Start();
    PWM8_1_WritePeriod(PWM_PERIOD);
    PWM8_1_WritePulseWidth(bpwm1);
    PWM8_1_EnableInt();

	pwm2 = bpwm2 = sine_calcwidth(sine_phase + SIN_TABLE_N3, amplitude);
	PWM8_2_WritePeriod((bpwm2>>1)-1);
    PWM8_2_Start();
    PWM8_2_WritePeriod(PWM_PERIOD);
    PWM8_2_WritePulseWidth(bpwm2);
    PWM8_2_EnableInt();

	pwm3 = bpwm3 = sine_calcwidth(sine_phase + (2*SIN_TABLE_N3), amplitude);
	PWM8_3_WritePeriod((bpwm3>>1)-1);
    PWM8_3_Start();
    PWM8_3_WritePeriod(PWM_PERIOD);
    PWM8_3_WritePulseWidth(bpwm3);
    PWM8_3_EnableInt();
	
	/* new settings are ok */
	pwm_load_flag = 0;
	
	/* Run All - Sync. Start */
    CMPPRG_1_SetRef(CMPPRG_1_REF0_062);
    M8C_EnableGInt;
    
    /* After this point, comparator may be reused for other purposes ...
       (do not forget to set pwm enable bits to high)
    */
}

void tfpwm_stop(void)
{
	PWM8_1_Stop();
	PWM8_1_DisableInt();
	PWM8_2_Stop();
	PWM8_2_DisableInt();
	PWM8_3_Stop();
	PWM8_3_DisableInt();
}

void tfpwm_set(unsigned char sine_phase, unsigned char amplitude)
{
	/* wait until pwm isr loads new settings */
	while(pwm_load_flag);

#ifdef DEBUG
	if (sine_phase==0) debug_sync_out = 0x08; else debug_sync_out = 0x00;
#endif
	
	/* calculate new values for the pwm */
	pwm1   >>= 1;
	period1  = PWM_PERIOD - pwm1;
	pwm1     = sine_calcwidth(sine_phase, amplitude);
	period1 += (pwm1 >> 1);

	pwm2   >>= 1;
	period2  = PWM_PERIOD - pwm2;
	pwm2     = sine_calcwidth(sine_phase + SIN_TABLE_N3, amplitude);
	period2 += (pwm2 >> 1);

	pwm3   >>= 1;
	period3  = PWM_PERIOD - pwm3;
	pwm3     = sine_calcwidth(sine_phase + (2*SIN_TABLE_N3), amplitude);
	period3 += (pwm3 >> 1);
	
	/* request pwm update */
	pwm_load_flag = 1;
}
